@@ -14,11 +14,8 @@
 #
 # You should have received a copy of the GNU Lesser General Public
 # License along with this program; if not, see <http://www.gnu.org/licenses/>.
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 3, or (at your option)
-# any later version.
 """High-level pipelines."""
+import contextlib
 import os
 
 from gi.repository import GES
@@ -26,7 +23,7 @@ from gi.repository import GLib
 from gi.repository import GObject
 from gi.repository import Gst
 
-from pitivi.check import videosink_factory
+from pitivi.check import VIDEOSINK_FACTORY
 from pitivi.utils.loggable import Loggable
 from pitivi.utils.misc import format_ns
 
@@ -74,7 +71,7 @@ class SimplePipeline(GObject.Object, Loggable):
 
     __gsignals__ = PIPELINE_SIGNALS
 
-    class RecoveryState(object):
+    class RecoveryState:
         NOT_RECOVERING = "not-recovering"
         STARTED_RECOVERING = "started-recovering"
         SEEKED_AFTER_RECOVERING = "seeked-after-recovering"
@@ -86,10 +83,10 @@ class SimplePipeline(GObject.Object, Loggable):
         self._pipeline = pipeline
         self._bus = self._pipeline.get_bus()
         self._bus.add_signal_watch()
-        self._bus.connect("message", self._busMessageCb)
+        self._bus.connect("message", self._bus_message_cb)
         self._listening = False  # for the position handler
-        self._listeningInterval = DEFAULT_POSITION_LISTENNING_INTERVAL
-        self._listeningSigId = 0
+        self._listening_interval = DEFAULT_POSITION_LISTENNING_INTERVAL
+        self._listening_sig_id = 0
         self._duration = Gst.CLOCK_TIME_NONE
         # The last known position.
         self._last_position = 0 * Gst.SECOND
@@ -100,23 +97,28 @@ class SimplePipeline(GObject.Object, Loggable):
         self._timeout_async_id = 0
         self._force_position_listener = False
 
-        self.video_sink = None
-        self.sink_widget = None
-
     def create_sink(self):
-        sink = Gst.ElementFactory.make(videosink_factory.get_name(), None)
-        self.sink_widget = sink.props.widget
+        """Creates a video sink and a widget for displaying it.
 
-        if videosink_factory.get_name() == "gtksink":
+        Returns:
+            (Gst.Element, Gtk.Widget): An element of type Gst.ElementFlags.SINK
+            and a widget connected to it.
+        """
+        factory_name = VIDEOSINK_FACTORY.get_name()
+        sink = Gst.ElementFactory.make(factory_name, None)
+        widget = sink.props.widget
+
+        if factory_name == "gtksink":
             self.info("Using gtksink")
-            self.video_sink = sink
+            video_sink = sink
         else:
-            self.info("Using glsinkbin around %s", videosink_factory.get_name())
-            sinkbin = Gst.ElementFactory.make("glsinkbin", None)
-            sinkbin.props.sink = sink
-            self.video_sink = sinkbin
+            self.info("Using glsinkbin around %s", VIDEOSINK_FACTORY.get_name())
+            video_sink = Gst.ElementFactory.make("glsinkbin", None)
+            video_sink.props.sink = sink
 
-    def setForcePositionListener(self, force):
+        return video_sink, widget
+
+    def set_force_position_listener(self, force):
         self._force_position_listener = force
 
     def release(self):
@@ -127,26 +129,25 @@ class SimplePipeline(GObject.Object, Loggable):
 
         The instance will no longer be usable.
         """
-        self._removeWaitingForAsyncDoneTimeout()
-        self.deactivatePositionListener()
-        self._bus.disconnect_by_func(self._busMessageCb)
+        self._remove_waiting_for_async_done_timeout()
+        self.deactivate_position_listener()
+        self._bus.disconnect_by_func(self._bus_message_cb)
         self._bus.remove_signal_watch()
 
         self._pipeline.set_state(Gst.State.NULL)
         self._bus = None
 
-    def flushSeek(self):
-        if self.getState() == Gst.State.PLAYING:
+    def flush_seek(self):
+        if self.get_simple_state() == Gst.State.PLAYING:
             self.debug("Playing, no need to flush here!")
             return
 
         try:
-            self.simple_seek(self.getPosition())
+            self.simple_seek(self.get_position())
         except PipelineError as e:
             self.warning("Could not flush because: %s", e)
-            pass
 
-    def setState(self, state):
+    def set_simple_state(self, state):
         """Sets the low-level pipeline to the specified state.
 
         Raises:
@@ -156,16 +157,16 @@ class SimplePipeline(GObject.Object, Loggable):
         """
         self.debug("Setting state to: %r", state)
         if state >= Gst.State.PAUSED:
-            cstate = self.getState()
+            cstate = self.get_simple_state()
             if cstate < Gst.State.PAUSED:
                 if cstate == Gst.State.NULL:
                     timeout = MAX_BRINGING_TO_PAUSED_DURATION
                 else:
                     timeout = MAX_SET_STATE_DURATION
 
-                self._addWaitingForAsyncDoneTimeout(timeout)
+                self._add_waiting_for_async_done_timeout("set_simple_state: %s" % state, timeout)
         else:
-            self._removeWaitingForAsyncDoneTimeout()
+            self._remove_waiting_for_async_done_timeout()
 
         res = self._pipeline.set_state(state)
         if res == Gst.StateChangeReturn.FAILURE:
@@ -174,7 +175,7 @@ class SimplePipeline(GObject.Object, Loggable):
             raise PipelineError(
                 "Failure changing state of the Gst.Pipeline to %r, currently reset to NULL" % state)
 
-    def getState(self):
+    def get_simple_state(self):
         """Queries the low-level pipeline for the current state.
 
         This will do an actual query to the underlying GStreamer Pipeline.
@@ -190,20 +191,20 @@ class SimplePipeline(GObject.Object, Loggable):
 
     def play(self):
         """Sets the state to Gst.State.PLAYING."""
-        self.setState(Gst.State.PLAYING)
+        self.set_simple_state(Gst.State.PLAYING)
 
     def pause(self):
         """Sets the state to Gst.State.PAUSED."""
-        self.setState(Gst.State.PAUSED)
+        self.set_simple_state(Gst.State.PAUSED)
 
     def stop(self):
         """Sets the state to Gst.State.READY."""
-        self.setState(Gst.State.READY)
+        self.set_simple_state(Gst.State.READY)
 
     def playing(self):
-        return self.getState() == Gst.State.PLAYING
+        return self.get_simple_state() == Gst.State.PLAYING
 
-    def togglePlayback(self):
+    def toggle_playback(self):
         if self.playing():
             self.pause()
         else:
@@ -211,7 +212,7 @@ class SimplePipeline(GObject.Object, Loggable):
 
     # Position and Seeking methods
 
-    def getPosition(self, fails=True):
+    def get_position(self, fails=True):
         """Gets the current position of the low-level pipeline.
 
         Returns:
@@ -223,8 +224,8 @@ class SimplePipeline(GObject.Object, Loggable):
         try:
             res, cur = self._pipeline.query_position(Gst.Format.TIME)
         except Exception as e:
-            self.handleException(e)
-            raise PipelineError("Couldn't get position")
+            self.handle_exception(e)
+            raise PipelineError("Couldn't get position") from e
 
         if res:
             self._last_position = cur
@@ -237,16 +238,16 @@ class SimplePipeline(GObject.Object, Loggable):
         self.log("Got position %s", format_ns(cur))
         return cur
 
-    def getDuration(self):
+    def get_duration(self):
         """Gets the duration of the low-level pipeline."""
-        dur = self._getDuration()
+        dur = self._get_duration()
         self.log("Got duration %s", format_ns(dur))
         if self._duration != dur:
             self.emit("duration-changed", dur)
         self._duration = dur
         return dur
 
-    def activatePositionListener(self, interval=DEFAULT_POSITION_LISTENNING_INTERVAL):
+    def activate_position_listener(self, interval=DEFAULT_POSITION_LISTENNING_INTERVAL):
         """Activates the position listener.
 
         When activated, the instance will emit the `position` signal at the
@@ -261,56 +262,59 @@ class SimplePipeline(GObject.Object, Loggable):
         if self._listening:
             return True
         self._listening = True
-        self._listeningInterval = interval
+        self._listening_interval = interval
         # if we're in playing, switch it on
-        self._listenToPosition(self.getState() == Gst.State.PLAYING)
+        self._listen_to_position(self.get_simple_state() == Gst.State.PLAYING)
         return True
 
-    def deactivatePositionListener(self):
+    def deactivate_position_listener(self):
         """De-activates the position listener."""
-        self._listenToPosition(False)
+        self._listen_to_position(False)
         self._listening = False
 
-    def _positionListenerCb(self):
+    def _position_listener_cb(self):
         try:
             try:
-                position = self.getPosition()
+                position = self.get_position()
             except PipelineError as e:
                 self.warning("Could not get position because: %s", e)
             else:
                 if position != Gst.CLOCK_TIME_NONE:
                     self.emit("position", position)
         finally:
-            return True
+            # Call me again.
+            return True  # pylint: disable=lost-exception
 
-    def _listenToPosition(self, listen=True):
+    def _listen_to_position(self, listen=True):
         # stupid and dumm method, not many checks done
         # i.e. it does NOT check for current state
         if listen:
-            if self._listening and self._listeningSigId == 0:
-                self._listeningSigId = GLib.timeout_add(
-                    self._listeningInterval,
-                    self._positionListenerCb)
-        elif self._listeningSigId != 0:
-            GLib.source_remove(self._listeningSigId)
-            self._listeningSigId = 0
+            if self._listening and self._listening_sig_id == 0:
+                self._listening_sig_id = GLib.timeout_add(
+                    self._listening_interval,
+                    self._position_listener_cb)
+        elif self._listening_sig_id != 0:
+            GLib.source_remove(self._listening_sig_id)
+            self._listening_sig_id = 0
 
-    def _async_done_not_received_cb(self):
-        self.error("we didn't get async done, this is a bug")
-        self._removeWaitingForAsyncDoneTimeout()
+    def _async_done_not_received_cb(self, reason, timeout):
+        self.error("Async operation timed out after %d seconds, aborting: %s", timeout, reason)
+        self._remove_waiting_for_async_done_timeout()
         self._recover()
         return False
 
-    def _removeWaitingForAsyncDoneTimeout(self):
+    def _remove_waiting_for_async_done_timeout(self):
         if not self._busy_async:
             return
+
         GLib.source_remove(self._timeout_async_id)
         self._timeout_async_id = 0
 
-    def _addWaitingForAsyncDoneTimeout(self, timeout=WATCHDOG_TIMEOUT):
-        self._removeWaitingForAsyncDoneTimeout()
+    def _add_waiting_for_async_done_timeout(self, reason, timeout=WATCHDOG_TIMEOUT):
+        self._remove_waiting_for_async_done_timeout()
         self._timeout_async_id = GLib.timeout_add_seconds(timeout,
-                                                          self._async_done_not_received_cb)
+                                                          self._async_done_not_received_cb,
+                                                          reason, timeout)
 
     @property
     def _busy_async(self):
@@ -340,7 +344,7 @@ class SimplePipeline(GObject.Object, Loggable):
         Raises:
             PipelineError: When the seek fails.
         """
-        if self._busy_async or self.getState() < Gst.State.PAUSED:
+        if self._busy_async or self.get_simple_state() < Gst.State.PAUSED:
             self._next_seek = position
             self.info("Setting next seek to %s", self._next_seek)
             return
@@ -348,7 +352,7 @@ class SimplePipeline(GObject.Object, Loggable):
         self._next_seek = None
 
         # clamp between [0, duration]
-        position = max(0, min(position, self.getDuration()))
+        position = max(0, min(position, self.get_duration()))
         self.debug("Seeking to position: %s", format_ns(position))
         res = self._pipeline.seek(1.0,
                                   Gst.Format.TIME,
@@ -360,21 +364,21 @@ class SimplePipeline(GObject.Object, Loggable):
         if not res:
             raise PipelineError(self.get_name() + " seek failed: " + str(position))
 
-        self._addWaitingForAsyncDoneTimeout()
+        self._add_waiting_for_async_done_timeout("simple_seek: %s" % position)
 
         self.emit('position', position)
 
-    def seekRelative(self, time_delta):
+    def seek_relative(self, time_delta):
         try:
-            self.simple_seek(self.getPosition() + int(time_delta))
+            self.simple_seek(self.get_position() + int(time_delta))
         except PipelineError:
             self.error("Error while seeking %s relative", time_delta)
 
     # Private methods
 
-    def _busMessageCb(self, unused_bus, message):
+    def _bus_message_cb(self, unused_bus, message):
         if message.type == Gst.MessageType.EOS:
-            self.__emitPosition()
+            self.__emit_position()
             self.pause()
             self.emit('eos')
         elif message.type == Gst.MessageType.STATE_CHANGED:
@@ -387,11 +391,10 @@ class SimplePipeline(GObject.Object, Loggable):
                 if prev == Gst.State.READY and new == Gst.State.PAUSED:
                     # trigger duration-changed
                     try:
-                        self.getDuration()
+                        self.get_duration()
                     except PipelineError as e:
                         self.warning("Could not get duration because: %s", e)
                         # no sinks??
-                        pass
 
                     if self._recovery_state == self.RecoveryState.STARTED_RECOVERING:
                         if self._attempted_recoveries == MAX_RECOVERIES:
@@ -406,47 +409,48 @@ class SimplePipeline(GObject.Object, Loggable):
                             if self._next_seek is not None:
                                 position = self._next_seek
                             self.simple_seek(position)
-                    self._listenToPosition(self._force_position_listener)
+                    self._listen_to_position(self._force_position_listener)
                 elif prev == Gst.State.PAUSED and new == Gst.State.PLAYING:
-                    self._listenToPosition(True)
+                    self._listen_to_position(True)
                 elif prev == Gst.State.PLAYING and new == Gst.State.PAUSED:
-                    self._listenToPosition(self._force_position_listener)
+                    self._listen_to_position(self._force_position_listener)
 
                 if emit_state_change:
                     self.emit('state-change', new, prev)
 
         elif message.type == Gst.MessageType.ERROR:
             error, detail = message.parse_error()
-            self._handleErrorMessage(error, detail, message.src)
+            self._handle_error_message(error, detail, message.src)
             Gst.debug_bin_to_dot_file_with_ts(self._pipeline,
                                               Gst.DebugGraphDetails.ALL,
                                               "pitivi.error")
-            if not self._rendering():
+            if not self.rendering():
+                self._remove_waiting_for_async_done_timeout()
                 self._recover()
         elif message.type == Gst.MessageType.DURATION_CHANGED:
-            self.debug("Duration might have changed, querying it")
-            GLib.idle_add(self._queryDurationAsync)
+            self.debug("Querying duration async, because it changed")
+            GLib.idle_add(self._query_duration_async)
         elif message.type == Gst.MessageType.ASYNC_DONE:
             self.debug("Async done, ready for action")
             self.emit("async-done")
-            self._removeWaitingForAsyncDoneTimeout()
+            self._remove_waiting_for_async_done_timeout()
             if self._recovery_state == self.RecoveryState.SEEKED_AFTER_RECOVERING:
                 self._recovery_state = self.RecoveryState.NOT_RECOVERING
                 self._attempted_recoveries = 0
-            self.__emitPosition()
+            self.__emit_position()
             if self._next_seek is not None:
                 self.info("Performing seek after ASYNC_DONE")
                 self.simple_seek(self._next_seek)
         else:
             self.log("%s [%r]", message.type, message.src)
 
-    def __emitPosition(self):
+    def __emit_position(self):
         # When the pipeline has been paused we need to update the
         # timeline/playhead position, as the 'position' signal
         # is only emitted every DEFAULT_POSITION_LISTENNING_INTERVAL
         # ms and the playhead jumps during the playback.
         try:
-            position = self.getPosition()
+            position = self.get_position()
         except PipelineError as e:
             self.warning("Getting the position failed: %s", e)
             return None
@@ -459,66 +463,76 @@ class SimplePipeline(GObject.Object, Loggable):
     def _recover(self):
         if not self._bus:
             raise PipelineError("Should not try to recover after destroy")
+
         if self._attempted_recoveries == MAX_RECOVERIES:
             self.emit("died")
-            self.error(
-                "Pipeline error detected multiple times in a row, not resetting anymore")
+            self.error("Declaring pipeline dead, because %d successive reset attempts failed", MAX_RECOVERIES)
             return
 
-        self.error("Pipeline error detected during playback, resetting"
-                   " -- num tries: %d", self._attempted_recoveries)
-
-        self.setState(Gst.State.NULL)
+        self._attempted_recoveries += 1
+        self.error("Resetting pipeline because error detected during playback. "
+                   "Try %d", self._attempted_recoveries)
+        self.set_simple_state(Gst.State.NULL)
         self._recovery_state = self.RecoveryState.STARTED_RECOVERING
         self.pause()
 
-        self._attempted_recoveries += 1
-
-    def _queryDurationAsync(self, *unused_args, **unused_kwargs):
+    def _query_duration_async(self, *unused_args, **unused_kwargs):
         try:
-            self.getDuration()
-        except Exception as e:
+            self.get_duration()
+        except PipelineError as e:
             self.warning("Could not get duration because: %s", e)
         return False
 
-    def _handleErrorMessage(self, error, detail, source):
+    def _handle_error_message(self, error, detail, source):
         self.error("error from %s: %s (%s)", source, error, detail)
         self.emit('error', error.message, detail)
 
-    def _getDuration(self):
+    def _get_duration(self):
         try:
             res, dur = self._pipeline.query_duration(Gst.Format.TIME)
         except Exception as e:
-            self.handleException(e)
-            raise PipelineError("Couldn't get duration: %s" % e)
+            self.handle_exception(e)
+            raise PipelineError("Couldn't get duration") from e
 
         if not res:
             raise PipelineError("Couldn't get duration: Returned None")
         return dur
 
-    def _rendering(self):
+    def rendering(self):
         return False
 
 
 class AssetPipeline(SimplePipeline):
-    """Pipeline for playing a single clip."""
+    """Pipeline for playing a single asset.
 
-    def __init__(self, clip=None, name=None):
+    Attributes:
+        uri (str): The low-level pipeline.
+    """
+
+    def __init__(self, uri=None, name=None):
         ges_pipeline = Gst.ElementFactory.make("playbin", name)
+        ges_pipeline.props.video_filter = Gst.parse_bin_from_description("videoflip method=automatic", True)
         SimplePipeline.__init__(self, ges_pipeline)
 
-        self.create_sink()
-
-        self.clip = clip
-        if self.clip:
-            self.setClipUri(self.clip.props.uri)
+        self.__uri = None
+        if uri:
+            self.uri = uri
 
     def create_sink(self):
-        SimplePipeline.create_sink(self)
-        self._pipeline.set_property("video_sink", self.video_sink)
+        video_sink, sink_widget = SimplePipeline.create_sink(self)
+        self._pipeline.set_property("video_sink", video_sink)
+        return video_sink, sink_widget
 
-    def setClipUri(self, uri):
+    @property
+    def uri(self):
+        # We could maybe get it using `self._pipeline.get_property`, but
+        # after setting the state to Gst.State.PAUSED, it becomes None.
+        return self.__uri
+
+    @uri.setter
+    def uri(self, uri):
         self._pipeline.set_property("uri", uri)
+        self.__uri = uri
 
 
 class Pipeline(GES.Pipeline, SimplePipeline):
@@ -534,6 +548,7 @@ class Pipeline(GES.Pipeline, SimplePipeline):
 
         self._was_empty = False
         self._commit_wanted = False
+        self._prevent_commits = 0
 
         if "watchdog" in os.environ.get("PITIVI_UNSTABLE_FEATURES", ''):
             watchdog = Gst.ElementFactory.make("watchdog", None)
@@ -545,33 +560,35 @@ class Pipeline(GES.Pipeline, SimplePipeline):
                 self.props.audio_filter = watchdog
 
     def create_sink(self):
-        SimplePipeline.create_sink(self)
-        self._pipeline.preview_set_video_sink(self.video_sink)
+        video_sink, sink_widget = SimplePipeline.create_sink(self)
+        self._pipeline.preview_set_video_sink(video_sink)
+        return video_sink, sink_widget
 
     def set_mode(self, mode):
         self._next_seek = None
         return GES.Pipeline.set_mode(self, mode)
 
-    def _getDuration(self):
+    def _get_duration(self):
         return self.props.timeline.get_duration()
 
     def do_change_state(self, state):
         if state == Gst.StateChange.PAUSED_TO_READY:
-            self._removeWaitingForAsyncDoneTimeout()
+            self._remove_waiting_for_async_done_timeout()
 
         return GES.Pipeline.do_change_state(self, state)
 
-    def stepFrame(self, framerate, frames_offset):
+    def step_frame(self, framerate, frames_offset):
         """Seeks backwards or forwards the specified amount of frames.
 
         This clamps the playhead to the project frames.
 
         Args:
-            frames_offsets (int): The number of frames to step. Negative number
+            framerate (Gst.Fraction): The framerate of the project.
+            frames_offset (int): The number of frames to step. Negative number
                 for stepping backwards.
         """
         try:
-            position = self.getPosition()
+            position = self.get_position()
         except PipelineError:
             self.warning(
                 "Couldn't get position (you're framestepping too quickly), ignoring this request")
@@ -594,13 +611,13 @@ class Pipeline(GES.Pipeline, SimplePipeline):
             # Nowhere to seek.
             return
 
-        if self._rendering():
+        if self.rendering():
             raise PipelineError("Trying to seek while rendering")
 
         st = Gst.Structure.new_empty("seek")
-        if self.getState() == Gst.State.PLAYING:
+        if self.get_simple_state() == Gst.State.PLAYING:
             st.set_value("playback_time", float(
-                self.getPosition()) / Gst.SECOND)
+                self.get_position()) / Gst.SECOND)
         st.set_value("start", float(position / Gst.SECOND))
         st.set_value("flags", "accurate+flush")
         self.app.write_action(st)
@@ -611,43 +628,53 @@ class Pipeline(GES.Pipeline, SimplePipeline):
             self.error("Error while seeking to position: %s, reason: %s",
                        format_ns(position), e)
 
-    def _busMessageCb(self, bus, message):
+    def _bus_message_cb(self, bus, message):
         if message.type == Gst.MessageType.ASYNC_DONE:
-            self.commiting = False
-            self.app.gui.timeline_ui.timeline.update_visible_overlays()
+            self.app.gui.editor.timeline_ui.timeline.update_visible_overlays()
 
         if message.type == Gst.MessageType.ASYNC_DONE and\
                 self._commit_wanted:
-            self.debug("Commiting now that ASYNC is DONE")
-            self._addWaitingForAsyncDoneTimeout()
+            self.debug("Committing now that ASYNC is DONE")
+            self._add_waiting_for_async_done_timeout("_bus_message_cb: committing")
             self.props.timeline.commit()
             self._commit_wanted = False
         else:
-            SimplePipeline._busMessageCb(self, bus, message)
+            SimplePipeline._bus_message_cb(self, bus, message)
+
+    @contextlib.contextmanager
+    def commit_timeline_after(self):
+        self._prevent_commits += 1
+        self.info("Disabling commits during action execution")
+        try:
+            yield
+        finally:
+            self._prevent_commits -= 1
+            self.commit_timeline()
 
     def commit_timeline(self):
-        if self.getState() == Gst.State.NULL:
+        if self._prevent_commits > 0 or self.get_simple_state() == Gst.State.NULL:
             # No need to commit. NLE will do it automatically when
             # changing state from READY to PAUSED.
             return
+
         is_empty = self.props.timeline.is_empty()
         if self._busy_async and not self._was_empty and not is_empty:
             self._commit_wanted = True
             self._was_empty = False
             self.log("commit wanted")
         else:
-            self._addWaitingForAsyncDoneTimeout()
+            self._add_waiting_for_async_done_timeout("commit_timeline")
             self.props.timeline.commit()
-            self.debug("Commiting right now")
+            self.debug("Committing right now")
             self._was_empty = is_empty
 
-    def setState(self, state):
-        SimplePipeline.setState(self, state)
+    def set_simple_state(self, state):
+        SimplePipeline.set_simple_state(self, state)
         if state >= Gst.State.PAUSED and self.props.timeline.is_empty():
-            self.debug("No ASYNC_DONE will be emited on empty timelines")
+            self.debug("No ASYNC_DONE will be emitted on empty timelines")
             self._was_empty = True
-            self._removeWaitingForAsyncDoneTimeout()
+            self._remove_waiting_for_async_done_timeout()
 
-    def _rendering(self):
+    def rendering(self):
         mask = GES.PipelineFlags.RENDER | GES.PipelineFlags.SMART_RENDER
         return self._pipeline.get_mode() & mask != 0
